import numpy as np
import xarray as xr

try:
    from best_estimate_common import BestEstimate
    from instrument_source import *
except ImportError:
    from .best_estimate_common import BestEstimate
    from .instrument_source import *

class BestEstimate2Instruments(BestEstimate):
    input_dataset_names = ['sirsE', 'sirsC']

    SOURCE_1 = 1
    SOURCE_2 = 2
    SOURCE_1_2 = 0

    def __init__(self):
        super().__init__()
        self.init_split_source_flag_dict()
        self.allocate_source_bit_lookup()

    def check_last_best_estimate(self, offset:int,
        last_best_estimate_source:int, data_available_flag:int) -> None: 
        if data_available_flag == self.SOURCE_1_2:
            self.data_flag_out[offset] = self.check_last_best_estimate_two_inst_both_available(
                last_best_estimate_source, self.SOURCE_1, self.SOURCE_2)
        else:
            self.data_flag_out[offset] = self.check_last_best_estimate_one_inst_available(
                last_best_estimate_source, self.data_flag_out[offset])

        # When flag is INSUFFICIENT_DATA, 3-instrument vars use the value anyway;
        # 2-instrument vars do not
        # SUGGESTED CHANGE: let the instrument_diff get set to MISSING here
        if self.data_flag_out[offset] == FLAG_INSUFFICIENT_DATA:
            instr_diff = self.instrument_diffs[offset]
            self.set_value_failed(offset)
            self.instrument_diffs[offset] = instr_diff

    def get_vals_this_offset(self, offset:int, key:str='data') -> float:
        return [self.input_data_dicts[varname][key][offset] 
            for varname in self.input_dataset_names] + [None]

    def set_val_lookup_table(self, lookup_table:dict, val1:float, val2:float, unused:None) -> None:
        lookup_table[self.SOURCE_1] = val1
        lookup_table[self.SOURCE_2] = val2

    def init_split_source_flag_dict(self) -> None:
        self.split_source_flag_dict = {
            self.SOURCE_1_2 : [self.SOURCE_1, self.SOURCE_2],
            self.SOURCE_1 : [self.SOURCE_1],
            self.SOURCE_2 : [self.SOURCE_2]
            
        }

        self.valid_data_flag_values = BestEstimate.get_valid_data_flag_values(
            self.split_source_flag_dict)

    def num_instruments_per_flag(self, data_flag:int) -> int:
        if data_flag == SOURCE_NONE or data_flag == FLAG_INSUFFICIENT_DATA:
            return 0
        if data_flag == self.SOURCE_1_2:
            return 2
        return 1

    def check_last_best_estimate_two_inst_both_available(self, last_best_estimate_source:int,
        source1:int, source2:int) -> int:
        if last_best_estimate_source == source1:
            return source1
        elif last_best_estimate_source == source2:
            return source2
        elif last_best_estimate_source in self.valid_data_flag_values:
            return FLAG_INSUFFICIENT_DATA
        else:
            return self.SOURCE_1_2

    def check_last_best_estimate_one_inst_available(self, last_best_estimate_source:int,
        available_source:int) -> int:
        if last_best_estimate_source not in self.valid_data_flag_values:
            return available_source
        if last_best_estimate_source != FLAG_INSUFFICIENT_DATA  and \
            available_source in self.split_source_flag_dict[last_best_estimate_source]:
            return available_source
        return FLAG_INSUFFICIENT_DATA
  
    def init_source_bit_lookup(self):
        sirsC_bit = BestEstimate.source_sirsC_bit
        sirsE_bit = BestEstimate.source_sirsE_bit

        self.source_bit_lookup[self.SOURCE_1] = sirsE_bit
        self.source_bit_lookup[self.SOURCE_2] = sirsC_bit
        self.source_bit_lookup[self.SOURCE_1_2] = sirsE_bit | sirsC_bit
