import numpy as np
import xarray as xr

from pdb import set_trace

try:
    from best_estimate_common import BestEstimate
    from instrument_source import *
except ImportError:
    from .best_estimate_common import BestEstimate
    from .instrument_source import *

class BestEstimate3Instruments(BestEstimate):
    input_dataset_names = ['sirsE', 'sirsC', 'brs']
    brs:xr.Dataset = None

    SOURCE_1 = -2
    SOURCE_2 = -3
    SOURCE_3 = -1
    SOURCE_1_2 = 2
    SOURCE_1_3 = 0
    SOURCE_2_3 = 1

    def __init__(self, varname:str, max_percent_diff:float):
        super().__init__()
        
        self.varname = varname
        self.max_diff = 5.0
        self.max_percent_diff = max_percent_diff
        self.apply_min_and_max_during_best_estimate = False

        self.init_split_source_flag_dict()
        self.allocate_source_bit_lookup()

    def get_input_fields(self) -> tuple:
        if self.input_data_dicts is None:
            self.input_data_dicts = {}
            self.input_data_dicts['sirsC'] = BestEstimate.get_data_qc_dict(
                self.sirsC, self.varname, 'sirsC')
            self.input_data_dicts['sirsE'] = BestEstimate.get_data_qc_dict(
                self.sirsE, self.varname, 'sirsE')
            self.input_data_dicts['brs'] = BestEstimate.get_data_qc_dict(
                self.brs, self.varname, 'brs')

        if self.data_flag_weekly is None:
            self.data_flag_weekly = self.get_weekly_flag(self.varname)

        return self.input_data_dicts, self.data_flag_weekly

    def get_vals_this_offset(self, offset:int, key:str='data') -> float:
        return [self.input_data_dicts[varname][key][offset] 
            for varname in self.input_dataset_names]

    def check_last_best_estimate(self, offset:int,
        last_best_estimate_source:int, data_available_flag:int) -> None: 
        if num_instruments_available(data_available_flag) == 1:
            self.data_flag_out[offset] = self.check_last_best_estimate_one_inst_available(
                last_best_estimate_source, self.data_flag_out[offset])
        else:
            if data_available_flag == AVAILABLE_ALL:
                source1, source2 = self.split_source_flag_dict[self.data_flag_out[offset]]
                self.data_flag_out[offset] = self.check_last_best_estimate_three_inst_all_available(
                    last_best_estimate_source, source1, source2)
            else:
                self.data_flag_out[offset] = self.check_last_best_estimate_three_inst_two_available(
                    last_best_estimate_source, self.data_flag_out[offset])


    def should_check_last_best_estimate(self, val_using_1:float,
        val_using_2:float, offset:int, data_available_flag:int) -> bool:
        result = False

        # If only one instrument is available:
        if val_using_2 is None:
            self.best_estimates[offset] = val_using_1
            self.instrument_diffs[offset] = self.missing_value
            result = True
        else:
            diff_vals_using = abs(val_using_1 - val_using_2)
            self.best_estimates[offset] = (val_using_1 + val_using_2) / 2
            self.instrument_diffs[offset] = diff_vals_using

            if not (diff_vals_using < self.max_diff or 
                (self.best_estimates[offset] != 0 and 
                diff_vals_using / self.best_estimates[offset] < self.max_percent_diff)):
                result = True

        return result

    def init_split_source_flag_dict(self) -> None:
        self.split_source_flag_dict = {
            self.SOURCE_1_2 : [self.SOURCE_1, self.SOURCE_2],
            self.SOURCE_1_3 : [self.SOURCE_1, self.SOURCE_3],
            self.SOURCE_2_3 : [self.SOURCE_2, self.SOURCE_3],
            self.SOURCE_1 : [self.SOURCE_1],
            self.SOURCE_2 : [self.SOURCE_2],
            self.SOURCE_3 : [self.SOURCE_3]
        }

        self.valid_data_flag_values = BestEstimate.get_valid_data_flag_values(
            self.split_source_flag_dict)

    def init_source_bit_lookup(self):
        sirsC_bit = BestEstimate.source_sirsC_bit
        sirsE_bit = BestEstimate.source_sirsE_bit
        brs_bit = BestEstimate.source_brs_bit

        self.source_bit_lookup[self.SOURCE_1] = sirsE_bit
        self.source_bit_lookup[self.SOURCE_2] = sirsC_bit
        self.source_bit_lookup[self.SOURCE_3] = brs_bit
        self.source_bit_lookup[self.SOURCE_1_2] = sirsE_bit | sirsC_bit
        self.source_bit_lookup[self.SOURCE_1_3] = sirsE_bit | brs_bit
        self.source_bit_lookup[self.SOURCE_2_3] = sirsC_bit | brs_bit

    def check_last_best_estimate_three_inst_two_available(self, last_best_estimate_source:int,
        flag_both_sources:int) -> int:
        source1, source2 = self.split_source_flag_dict[flag_both_sources]
        if last_best_estimate_source == flag_both_sources:
            return flag_both_sources
        elif last_best_estimate_source == source1:
            return source1
        elif last_best_estimate_source == source2:
            return source2
        return FLAG_INSUFFICIENT_DATA

    def check_last_best_estimate_three_inst_all_available(self, last_best_estimate_source:int,
        source1:int, source2:int) -> int:
        if last_best_estimate_source == source1:
            return source1
        elif last_best_estimate_source == source2:
            return source2
        return FLAG_INSUFFICIENT_DATA

    def check_last_best_estimate_one_inst_available(self, last_best_estimate_source:int,
        available_source:int) -> int:
        return available_source

    def emulate_invalid_data_flag_bug(self, offset:int, 
        data_available_flag:int) -> None:
        if data_available_flag == AVAILABLE_ALL:
            return

        self.best_estimates[offset] = 0
        self.data_flag_out[offset] = data_available_flag
        self.instrument_diffs[offset] = 0
        # The C VAP sometimes sets data_flag before checking weekly_flag
        if data_available_flag == AVAILABLE_1 and offset > 0:
            self.data_flag_out[offset] = self.SOURCE_1

    def num_instruments_per_flag(self, data_flag:int) -> int:
        if data_flag == SOURCE_NONE or data_flag == FLAG_INSUFFICIENT_DATA:
            return 0
        if data_flag == self.SOURCE_1 or data_flag == self.SOURCE_2 or data_flag == self.SOURCE_3:
            return 1
        return 2