import numpy as np
import xarray as xr
from pdb import set_trace
import dsproc3 as dsproc

from adi_py import SkipProcessingIntervalException

try:
    from BestEstimate3Instruments import BestEstimate3Instruments
    from instrument_source import *
    from var_names import down_short_diffuse_hemisp
    from best_estimate_common import BestEstimate
except ImportError:
    from .BestEstimate3Instruments import BestEstimate3Instruments
    from .instrument_source import *
    from .var_names import down_short_diffuse_hemisp
    from .best_estimate_common import BestEstimate

class BestEstimateDiffuse(BestEstimate3Instruments):
    sirsc1dutt = None
    sirse1dutt = None
    sirosc1dutt = None
    sirose1dutt = None
    brs1dutt = None
    brsn1dutt = None

    source_to_bit = {}
    input_data_dicts = None

    def __init__(self):
        self.varname = down_short_diffuse_hemisp
        self.max_percent_diff = 0.1
        self.max_diff = 0

        super().__init__(self.varname, self.max_percent_diff)

    def set_datasets(sirosc1dutt:xr.Dataset, sirose1dutt:xr.Dataset,
        sirs1dutt:xr.Dataset, sirse1dutt:xr.Dataset, brs1dutt:xr.Dataset,
        brsn1dutt:xr.Dataset) -> None:
        BestEstimateDiffuse.sirsc1dutt = sirs1dutt
        BestEstimateDiffuse.sirse1dutt = sirse1dutt
        BestEstimateDiffuse.sirosc1dutt = sirosc1dutt
        BestEstimateDiffuse.sirose1dutt = sirose1dutt
        BestEstimateDiffuse.brs1dutt = brs1dutt
        BestEstimateDiffuse.brsn1dutt = brsn1dutt

    def free_unneeded_input_data(self) -> None:
        super().free_unneeded_input_data()
        del BestEstimateDiffuse.sirsc1dutt, BestEstimateDiffuse.sirse1dutt, \
            BestEstimateDiffuse.sirose1dutt, BestEstimateDiffuse.sirosc1dutt, \
            BestEstimateDiffuse.brs1dutt, BestEstimateDiffuse.brsn1dutt

    def get_diffuse_corrected(self, datasets:list, dataset_suffix:str,
        num_candidates:int=3) -> np.ndarray:
        result = np.copy(self.placeholder_var_data['data'])
        qc_result = np.copy(self.placeholder_var_data['qc'])
        sources = ''
        data_bad_mask = np.full(len(qc_result), True)

        # For non-None dataset in datasets
        # Grab data and QC and overlay the good values on using mask
        # At end of each loop, check if there are any bad QC values
        # If so, move to next variable/dataset candidate
        for dataset in datasets:
            if dataset is None:
                continue

            for i in range(1, num_candidates + 1):
                varname = f'diffuse_{dataset_suffix}_{i}'
                try:
                    var_data = dataset[varname].data
                    if BestEstimate.begin_input > BestEstimate.retrieve_qc_vars_after:
                        qc_input = dataset['qc_' + varname].data
                    else:
                        qc_input = np.zeros(var_data.shape)
                        qc_input[var_data < BestEstimate.missing_value + 1] = \
                            BestEstimate.qc_missing_value

                    retrieved_var = dsproc.get_retrieved_var(varname, 0)
                    source_this_var = dsproc.get_source_ds_name(retrieved_var) +\
                        ':' + dsproc.get_source_var_name(retrieved_var) 
                except KeyError:
                    continue

                try:
                    input_begin_index, input_end_index, replace_through, replace_from, input_gaps =\
                        BestEstimate.get_gap_data(dataset)
                except IndexError as e:
                    raise SkipProcessingIntervalException(
                        f'Error reading in data for {varname}: {str(e)}')

                var_data = BestEstimate.fill_gaps(var_data, input_begin_index,
                    input_end_index, input_gaps)
                qc_input = BestEstimate.fill_gaps(qc_input, input_begin_index,
                    input_end_index, input_gaps, 
                    placeholder=BestEstimate.qc_missing_value)

                BestEstimate.conform_to_input_start_end_times(var_data, 
                    replace_through, replace_from)
                BestEstimate.conform_to_input_start_end_times(qc_input, 
                    replace_through, replace_from, 
                    placeholder=BestEstimate.qc_missing_value)

                if not np.any(var_data[data_bad_mask] > BestEstimate.missing_value):
                    continue

                input_bad_mask = np.asarray(qc_input > 0)
                sources += source_this_var

                result[data_bad_mask] = var_data[data_bad_mask]
                qc_result[data_bad_mask] = qc_input[data_bad_mask]

                data_bad_mask = np.logical_and(data_bad_mask, input_bad_mask)

                if not np.any(data_bad_mask):
                    return result, qc_result, sources

        return result, qc_result, sources

    def get_input_fields(self) -> tuple:
        if self.input_data_dicts is None:
            diffuse_sirsC, qc_diffuse_sirsC, sirsC_sources = self.get_diffuse_corrected(
                [self.sirsc1dutt, self.sirosc1dutt, self.sirsC], 'sirsC')
            diffuse_sirsE, qc_diffuse_sirsE, sirsE_sources = self.get_diffuse_corrected(
                [self.sirse1dutt, self.sirose1dutt, self.sirsE], 'E13')
            diffuse_brs, qc_diffuse_brs, brs_sources = self.get_diffuse_corrected(
                [self.brs1dutt, self.brsn1dutt, self.brs], 'brs')

            self.input_data_dicts = {
                'sirsC' : { 'data' : diffuse_sirsC, 'qc' : qc_diffuse_sirsC, 
                    'suffix' : '', 'source' : sirsC_sources },
                'sirsE' : { 'data' : diffuse_sirsE, 'qc' : qc_diffuse_sirsE, 
                    'suffix' : '_E13', 'source' : sirsE_sources },
                'brs' : { 'data' : diffuse_brs, 'qc' : qc_diffuse_brs,
                    'suffix' : "brs", 'source' : brs_sources }
            }

            # We need to save the inputs in a global location
            # so the quicklook hook can see them later
            BestEstimateDiffuse.input_data_dicts = self.input_data_dicts

        if self.data_flag_weekly is None:
            self.data_flag_weekly = self.get_weekly_flag(self.varname)

        return self.input_data_dicts, self.data_flag_weekly