import numpy as np
import xarray as xr
from pdb import set_trace
from adi_py import ADILogger

try:
    from .BestEstimate2Instruments import BestEstimate2Instruments, BestEstimate
    from .var_names import up_long_hemisp, surface_temp
except ImportError:
    from BestEstimate2Instruments import BestEstimate2Instruments, BestEstimate
    from var_names import up_long_hemisp, surface_temp

class BestEstimateUpLongHemisp(BestEstimate2Instruments):
    RUNNING_SDEV_MINUTES = 15

    def __init__(self, irt10m:xr.Dataset):
        super().__init__()

        self.irt10m = irt10m
        self.max_percent_diff = 0.04
        self.max_sdev = 0.01
        self.varname = up_long_hemisp

    def get_input_fields(self) -> tuple:
        if self.input_data_dicts is None:
            self.input_data_dicts = {}
            self.input_data_dicts['sirsC'] = BestEstimate.get_data_qc_dict(
                self.sirsC, up_long_hemisp, 'sirsC')
            self.input_data_dicts['sirsE'] = BestEstimate.get_data_qc_dict(
                self.sirsE, up_long_hemisp, 'sirsE')
            self.input_data_dicts['irt10m'] = BestEstimate.get_data_qc_dict(
                self.irt10m, surface_temp, 'irt10m')

        if self.data_flag_weekly is None:
            self.data_flag_weekly = self.get_weekly_flag(up_long_hemisp)

        return self.input_data_dicts, self.data_flag_weekly

    def should_check_last_best_estimate(self, val1:float, val2:float, offset:int,
        data_available_flag:int) -> bool:
        result = False

        # if both instruments are available
        if val2 is not None:
            self.best_estimates[offset] = (val1 + val2) / 2
            self.instrument_diffs[offset] = val1 - val2
            # SUGGESTED CHANGE: instrument_diff should be the abs of the diff,
            # like the 3-inst vars do
            if (self.best_estimates[offset] == 0 or
                abs((val2 - val1) / self.best_estimates[offset]) < self.max_percent_diff):
                self.data_flag_out[offset] = self.SOURCE_1_2
            else:
                sdev_half_length = BestEstimateUpLongHemisp.RUNNING_SDEV_MINUTES // 2
                begin_running_sdev = max(offset - sdev_half_length, 0)
                end_running_sdev = min(offset + sdev_half_length, BestEstimate.len_input_arrays)

                sirsC = self.input_data_dicts['sirsC']
                sirsE = self.input_data_dicts['sirsE']
                
                sdev_sirsC_ratio = self.running_sdev_surface_temp_over_flux(
                    begin_running_sdev, end_running_sdev, sirsC)
                sdev_sirsE_ratio = self.running_sdev_surface_temp_over_flux(
                    begin_running_sdev, end_running_sdev, sirsE)

                ADILogger.debug(f"offset: {offset}, sdev_sirsC_ratio: {sdev_sirsC_ratio}, sdev_sirsE_ratio: {sdev_sirsE_ratio}")

                if sdev_sirsC_ratio is None and sdev_sirsE_ratio is None:
                    result = True
                elif sdev_sirsC_ratio is None:
                    self.data_flag_out[offset] = self.SOURCE_1
                elif sdev_sirsE_ratio is None:
                    self.data_flag_out[offset] = self.SOURCE_2
                else:
                    if sdev_sirsC_ratio < self.max_sdev and sdev_sirsC_ratio < sdev_sirsE_ratio:
                        self.data_flag_out[offset] = self.SOURCE_2
                    elif sdev_sirsE_ratio < self.max_sdev and sdev_sirsE_ratio < sdev_sirsC_ratio:
                        self.data_flag_out[offset] = self.SOURCE_1
                    else:
                        result = True
        else:
            result = True

        return result

    def get_ratio_mask(qc_numerator:np.ndarray,
        denominator:np.ndarray, qc_denominator:np.ndarray,
        qc_missing_value:int) -> np.ndarray:
        numerator_good_mask = np.asarray(qc_numerator & qc_missing_value == 0)
        # Prevent divide by zero
        denominator_good_mask = np.asarray(qc_denominator & qc_missing_value == 0) & \
            np.asarray(denominator != 0)
        ratio_mask = np.logical_and(numerator_good_mask, denominator_good_mask)

        if np.any(ratio_mask):
            return ratio_mask
        return None

    # Takes dicts with keys "data" and "qc"
    # Call this once for sirsC and again for sirsE
    def running_sdev_surface_temp_over_flux(self, start_index:int, 
        end_index:int, flux:dict) -> float:
        surface_temp_data = self.input_data_dicts['irt10m']

        irt_slice = np.concatenate((surface_temp_data['data'][start_index:start_index+1], 
            np.zeros(end_index - start_index - 1, dtype=np.int32)))
        qc_irt_slice = np.concatenate((surface_temp_data['qc'][start_index:start_index+1], 
            np.zeros(end_index - start_index - 1, dtype=np.int32)))
        IRT = 5.67E-8 * np.power(irt_slice, 4)

        flux_slice = flux['data'][start_index:end_index]
        qc_flux_slice = flux['qc'][start_index:end_index]

        ratio_mask = BestEstimateUpLongHemisp.get_ratio_mask(
            qc_irt_slice, flux_slice, qc_flux_slice,
            self.qc_missing_value)

        if ratio_mask is None:
            return None

        return np.std(IRT[ratio_mask] / flux_slice[ratio_mask])

    def free_unneeded_input_data(self) -> None:
        super().free_unneeded_input_data()
        del self.irt10m, self.input_data_dicts['irt10m']

    # Does not override BestEstimate2Instruments::check_last_best_estimate