# **********************************************************************************************************************
#  Author:
#     name:  None
#     phone: None
#     email: None
# **********************************************************************************************************************

import numpy as np
import sys
import xarray as xr
import pandas
from datetime import date

from pdb import set_trace

import dsproc3 as dsproc
from adi_py import Process, ADILogger, SkipProcessingIntervalException
from adi_py.constants import ADIDatasetType
from .version import __version__

try:
    from .solar_position import solar_position
    from .var_names import *
    from .best_estimate_types import *
    from .calculate_derived_vars import *
    from .save_raw_data import save_raw_data
    from .make_plots import beflux1long_plots
except ImportError:
    from solar_position import solar_position
    from var_names import *
    from best_estimate_types import *
    from calculate_derived_vars import *
    from save_raw_data import save_raw_data
    from make_plots import beflux1long_plots

class Beflux1long(Process):
    """-----------------------------------------------------------------------------------------------------------------
    This class implements the beflux1long process using the new, XArray-based ADI library.  Please see
    https://adi-python.readthedocs.io/en/latest/examples.html for more examples on how to use this API.
    -----------------------------------------------------------------------------------------------------------------"""

    MISSING_VALUE = -9999

    # Input datastreams
    DS_IN_BEFLUX1LONG_C1 = 'beflux1long.c1'
    DS_IN_BRS_A2 = 'brs.a2'
    DS_IN_BRS_B1 = 'brs.b1'
    DS_IN_BRS1DUTT_C1 = 'brs1dutt.c1'
    DS_IN_BSRN_A1 = 'bsrn.a1'
    DS_IN_BSRN1DUTT_C1 = 'bsrn1dutt.c1'
    DS_IN_IRT10M_B1 = 'irt10m.b1'
    DS_IN_SIROS_A1 = 'siros.a1'
    DS_IN_SIROS1DUTT_C1 = 'siros1dutt.c1'
    DS_IN_SIRS_A1 = 'sirs.a1'
    DS_IN_SIRS_A2 = 'sirs.a2'
    DS_IN_SIRS_B1 = 'sirs.b1'
    DS_IN_SIRS1DUTT_C1 = 'sirs1dutt.c1'
    
    # Output datastreams
    DS_OUT_BEFLUX1LONG_C1 = 'beflux1long.c1'
    DS_OUT_QCFLUX1LONG_C1 = 'qcflux1long.c1'

    MISSING_VALUE = -9999
    QC_MISSING_VALUE = 1

    save_raw_command = 'W'
    dsproc.setopt(save_raw_command, None, None, 
        "Create CSV of input data and exit.")

    def get_product_name(ds_name:str, site:str, facility:str) -> str:
        data_level_index = ds_name.rfind('.')
        return site + ds_name[:data_level_index] + facility + \
            ds_name[data_level_index:]

    def epoch_to_yyyymmdd(timestamp:int) -> str:
        return date.fromtimestamp(timestamp).strftime("%Y%m%d")
    
    # platform_list: sorted from highest to lowest priority
    # used in situations where ADI only retrieved one
    #
    # The original code checked every single one,
    # and preferred the A2 over the B1 if it had more observations;
    # possible to emulate this? desirable?
    def get_dataset_from_list(self, platform_list:list, facility:str=None) -> xr.Dataset:
        for platform_name in platform_list:
            ds = self.get_retrieved_dataset(platform_name, facility=facility)
            if ds is not None:
                return ds
        return None

    # Assumes that BestEstimate's time interval vars
    # were already set
    def init_and_get_output_datasets(self, sample_times:list) -> xr.Dataset:
        time_out = sample_times

        # if we don't specify dataset_type,
        # it'll give us the ID of the INPUT beflux
        # and everything will break
        beflux1long_out = dsproc.get_output_dataset(
            self.get_dsid(self.DS_OUT_BEFLUX1LONG_C1,
            dataset_type=ADIDatasetType.OUTPUT))
        qcflux1long_out = dsproc.get_output_dataset(
            self.get_dsid(self.DS_OUT_QCFLUX1LONG_C1))

        if beflux1long_out is None or qcflux1long_out is None:
            error_message = "Failed to init the following dataset(s): "
            if not beflux1long_out:
                error_message += self.DS_OUT_BEFLUX1LONG_C1 + " "
            if not qcflux1long_out:
                error_message += self.DS_OUT_QCFLUX1LONG_C1 

            raise Exception(error_message)

        dsproc.set_sample_times(beflux1long_out, 0, time_out)
        dsproc.set_sample_times(qcflux1long_out, 0, time_out)

        return self.get_output_dataset(self.DS_OUT_BEFLUX1LONG_C1), \
            self.get_output_dataset(self.DS_OUT_QCFLUX1LONG_C1)

    # A temporary fix for data from April 24,2000 to May 9, 2000 where
    # two wires were crossed for SIRS C1
    def correct_known_bad_sirsC_data(sirsC:xr.Dataset, begin_date:int) -> None:
        if sirsC is None:
            return
        run_date = date.fromtimestamp(begin_date)
        april_24 = date(2000, 4, 24)
        may_2 = date(2000, 5, 2)
        may_9 = date(2000, 5, 9)

        if run_date < april_24 or run_date > may_9:
            return

        sirsC_dir = sirsC[short_direct_normal].data
        sirsC_diffuse = sirsC['diffuse_sirsC_3'].data

        if run_date == april_24:
            sirsC_dir = sirsC_dir[1154:]
            sirsC_diffuse = sirsC_dir[1154:]
        elif run_date == may_9:
            sirsC_dir = sirsC_dir[:1043]
            sirsC_diffuse = sirsC_dir[:1043]
        sirsC_dir[sirsC_dir > Beflux1long.MISSING_VALUE] *= -1
        sirsC_diffuse[sirsC_diffuse > Beflux1long.MISSING_VALUE] *= -1

        if run_date >= may_2:
            sirsC_up_short = sirsC[up_short_hemisp].data
            if run_date == may_2:
                sirsC_up_short = sirsC_up_short[908:]
            elif run_date == may_9:
                sirsC_up_short = sirsC_up_short[:1043]
            sirsC_up_short[sirsC_up_short > Beflux1long.MISSING_VALUE] *= -1
    
    def __init__(self):
        """-------------------------------------------------------------------------------------------------------------
        Class constructor for Beflux1long.  This method is used to initialize user data variables and main process
        settings that are stored within this class.  It is invoked BEFORE the dsproc main loop, so DO NOT call any
        init_process_hook methods here or you will likely get a segmentation fault error.
        -------------------------------------------------------------------------------------------------------------"""
        super().__init__()
        self._process_names = ['beflux1long']
        self._process_model = dsproc.PM_TRANSFORM_VAP
        self._process_version = __version__
        self._rollup_qc = True            # Roll up the qc bits of all vars with transformed qc bits in the output DOD
        self._include_debug_dumps = True  # This will automatically dump files if debug level > 1

        self.save_raw = False
        self.create_quicklooks = True

    def init_process_hook(self):
        self.save_raw = dsproc.getopt(Beflux1long.save_raw_command) is not None
        self.create_quicklooks = dsproc.getopt('--no-quicklook') is None
        

    def pre_transform_hook(self, begin_date: int, end_date: int):
        pass
        
    def post_transform_hook(self, begin_date: int, end_date: int): 
        pass     

    def process_data_hook(self, begin_date: int, end_date: int):
        latitude = 36.605
        longitude = -97.485
        altitude = 318

        # Retrieve our datasets
        sirsC = self.get_dataset_from_list(
            [self.DS_IN_SIRS_B1, self.DS_IN_SIRS_A2, self.DS_IN_SIRS_A1, self.DS_IN_SIROS_A1],
            facility='C1')
        sirsE = self.get_dataset_from_list(
            [self.DS_IN_SIRS_B1, self.DS_IN_SIRS_A2, self.DS_IN_SIRS_A1, self.DS_IN_SIROS_A1],
            facility='E13')
        brs = self.get_dataset_from_list(
            [self.DS_IN_BRS_B1, self.DS_IN_BRS_A2, self.DS_IN_BSRN_A1])
        sirsc1dutt = self.get_dataset_from_list(
            [self.DS_IN_SIRS1DUTT_C1], facility='C1')
        sirosc1dutt = self.get_dataset_from_list(
            [self.DS_IN_SIROS1DUTT_C1], facility='C1')
        sirse1dutt = self.get_dataset_from_list(
            [self.DS_IN_SIRS1DUTT_C1], facility='E13')
        sirose1dutt = self.get_dataset_from_list(
            [self.DS_IN_SIROS1DUTT_C1], facility='E13')
        brs1dutt = self.get_dataset_from_list([self.DS_IN_BRS1DUTT_C1])
        brsn1dutt = self.get_dataset_from_list([self.DS_IN_BSRN1DUTT_C1])

        # This order is important for set_output_start_and_end_timestamps
        possible_time_interval_models = [sirsE, sirsC, brs, sirose1dutt, 
            sirse1dutt, sirosc1dutt, sirsc1dutt, brsn1dutt, brs1dutt]

        output_sample_times = [begin_date + i * 60 for i in range(1440)]

        # Because no input vars appear in the output datasets,
        # we need to set output sample times manually
        beflux1long_out, qcflux1long_out = \
            self.init_and_get_output_datasets(output_sample_times)

        if beflux1long_out is None:
            raise Exception(f'Could not find output datastream for { self.DS_OUT_BEFLUX1LONG_C1 }')
        if qcflux1long_out is None:
            raise Exception(f'Could not find output datastream for { self.DS_OUT_QCFLUX1LONG_C1 }')

        no_input_data = not any(possible_time_interval_models)

        if no_input_data:
            raise SkipProcessingIntervalException("No SIRS or BRS data for this time window")

        irt10m = self.get_retrieved_dataset(self.DS_IN_IRT10M_B1)
        # Retrieves data for this day
        # and up to 1 week previous
        beflux1long_in = self.get_retrieved_dataset(self.DS_IN_BEFLUX1LONG_C1)

        Beflux1long.correct_known_bad_sirsC_data(sirsC, begin_date)
        BestEstimate.set_input_data_vars(begin_date, end_date)
        BestEstimate.set_input_start_and_end_timestamps(possible_time_interval_models)
        BestEstimate.set_datasets(sirsC, sirsE, beflux1long_in)
        BestEstimate3Instruments.brs = brs
        BestEstimateDiffuse.set_datasets(sirosc1dutt, sirose1dutt,
            sirsc1dutt, sirse1dutt, brs1dutt, brsn1dutt)
        del sirsC, sirsE, beflux1long_in, brs, sirsc1dutt, sirosc1dutt, \
            sirse1dutt, sirose1dutt, brs1dutt, brsn1dutt

        # Get zenith angle
        one_minute = np.timedelta64(1, 'm').astype('timedelta64[ns]')
        time_out = beflux1long_out['time'].data
        extra_times_before = [time_out[0] - i * one_minute
            for i in range(BestEstimate.EXTRA_MINUTES_RETRIEVED, 0, -1)]
        extra_times_after = [time_out[-1] + i * one_minute
            for i in range(BestEstimate.EXTRA_MINUTES_RETRIEVED)]
        time_in = np.concatenate((
            np.array(extra_times_before), time_out, np.array(extra_times_after)
        ))

        zenith_degrees = np.ndarray(time_in.shape)
        solar_distance = np.ndarray(time_in.shape)
        for i in range(len(time_in)):
            timestamp = pandas.Timestamp(time_in[i])
            solar_altitude, solar_distance[i] = solar_position(timestamp,
                latitude, longitude)
            zenith_degrees[i] = 90 - solar_altitude
        del time_out, extra_times_before, extra_times_after, one_minute

        # Calculate our best estimates
        best_estimate_objects = {
            short_direct_normal : BestEstimate3Instruments(short_direct_normal, 0.05),
            down_long_hemisp : BestEstimate3Instruments(down_long_hemisp, 0.2),
            down_short_diffuse_hemisp : BestEstimateDiffuse(),
            up_long_hemisp : BestEstimateUpLongHemisp(irt10m),
            up_short_hemisp : BestEstimateUpShortHemisp()
        }
        best_estimate_objects[up_short_hemisp].set_qc_thresholds(zenith_degrees)

        # Alternatively, save input netcdf files as raw data and exit
        if self.save_raw:
            message = save_raw_data(best_estimate_objects, time_in, zenith_degrees, begin_date)
            raise SkipProcessingIntervalException(message + "; continuing")

        for varname in best_estimate_objects:
            best_estimate_objects[varname].calc_best_estimate()

        # Need to be sure we clear this between processing intervals
        BestEstimate.dataset_gap_info = {}
        del BestEstimate.beflux1long_in

        # down_short_hemisp doesn't work
        # unless it's run after short and diffuse are calculated
        zenith_radians = np.radians(zenith_degrees)
        best_estimate_down_short = BestEstimateDownShortHemisp(
            best_estimate_objects[short_direct_normal].best_estimates,
            best_estimate_objects[short_direct_normal].qc_best_estimates,
            best_estimate_objects[down_short_diffuse_hemisp].best_estimates,
            best_estimate_objects[down_short_diffuse_hemisp].qc_best_estimates,
            zenith_radians)
        best_estimate_down_short.calc_best_estimate()
        best_estimate_objects[down_short_hemisp] = best_estimate_down_short
        del best_estimate_down_short, irt10m

        for varname in best_estimate_objects:
            best_estimate_obj = best_estimate_objects[varname]
            # This function will return None for down_short_hemisp
            # because it doesn't have a qc diff var
            if best_estimate_obj.calc_best_estimate_diffs() is not None:
                best_estimate_obj.set_best_estimate_diffs(qcflux1long_out)

        # Test data for physically impossible and extremely rare limits
        excessive_zenith_mask = np.asarray(zenith_degrees > 90)
        mu0 = np.cos(zenith_radians)
        mu0[excessive_zenith_mask] = 0
        Sa = 1367 / np.power(solar_distance, 2)
        a1 = Sa * mu0 ** 1.2
        del solar_distance, zenith_radians

        var_limits = [
            { 'varname' : down_short_hemisp, 
            'max' : 1.2 * a1 + 50, 'min' : -2},
            { 'varname' : down_short_diffuse_hemisp, 
            'max' : 0.75 * a1 + 30, 'min' : -2},
            { 'varname' : short_direct_normal, 
            'max' : 0.95 * Sa * np.power(mu0, 0.2) + 10, 'min' : -2},
            { 'varname' : up_short_hemisp, 
            'max' : a1 + 50, 'min' : -2},
            { 'varname' : down_long_hemisp, 'max' : 500, 'min' : 60},
            { 'varname' : up_long_hemisp, 'max' : 700, 'min' : 60}
        ]

        for limit_info in var_limits:
            best_estimate_obj = best_estimate_objects[limit_info['varname']]
            best_estimate_obj.set_min_and_max(limit_info['min'], limit_info['max'])
            best_estimate_obj.apply_min_and_max()
            best_estimate_obj.set_output_vars(beflux1long_out)
            best_estimate_obj.set_best_estimate_only(qcflux1long_out)

        up_short_hemisp_data = best_estimate_objects[up_short_hemisp].best_estimates
        down_short_hemisp_data = best_estimate_objects[down_short_hemisp].best_estimates
        up_long_hemisp_data = best_estimate_objects[up_long_hemisp].best_estimates
        down_long_hemisp_data = best_estimate_objects[down_long_hemisp].best_estimates

        # The calculation functions also handle mapping to the output interval
        zenith_out = BestEstimate.to_output_interval(zenith_degrees)
        net_surface_radiation_out = calc_net_surface_radiation(
            up_short_hemisp_data, down_short_hemisp_data, up_long_hemisp_data,
            down_long_hemisp_data)
        albedo_out = calc_albedo(up_short_hemisp_data,
            down_short_hemisp_data, zenith_degrees)

        for ds_out in [beflux1long_out, qcflux1long_out]:
            ds_out[zenith].data = zenith_out
            ds_out[net_surface_radiation].data = net_surface_radiation_out
            ds_out[albedo].data = albedo_out
         
        # Now you need to push any changes you made back to ADI
        self.sync_datasets(beflux1long_out)
        self.sync_datasets(qcflux1long_out)
      
    def finish_process_hook(self):
        """-----------------------------------------------------------------------
        This hook will be called once just after the main data processing loop finishes.  This function should be used
        to clean up any temporary files used.

        -----------------------------------------------------------------------"""
        pass

    def get_files_to_plot(ds_name:str, site:str, facility:str, 
        begin_date_string:str, end_date_string:str,
        output_dir:str) -> list:
        from os import listdir

        product_name = Beflux1long.get_product_name(ds_name, site, facility)
        files_dir = f'{output_dir}/{site}/{product_name}'

        files = []
        for filename in listdir(files_dir):
            filename_parts = filename.split('.')
            if len(filename_parts) < 3:
                continue

            datestamp = filename_parts[-3]
            if datestamp >= begin_date_string and datestamp < end_date_string:
                files.append(f'{files_dir}/{filename}')

        return files

    def quicklook_hook(self, begin_date: int, end_date: int):
        if not self.create_quicklooks:
            return
        from os import getenv, makedirs
        from os.path import isdir

        # If we ever run this process at a facility other than sgp C1,
        # we'll have to revisit this
        site = 'sgp'
        facility = 'C1'

        datastream_data_out = getenv('DATASTREAM_DATA_OUT')
        if datastream_data_out is None:
            datastream_data_out = getenv('DATASTREAM_DATA')

        begin_date_string = Beflux1long.epoch_to_yyyymmdd(begin_date)
        end_date_string = Beflux1long.epoch_to_yyyymmdd(end_date)

        beflux_files = Beflux1long.get_files_to_plot(self.DS_OUT_BEFLUX1LONG_C1,
            site, facility, begin_date_string, end_date_string, datastream_data_out)
        qcflux_files = Beflux1long.get_files_to_plot(self.DS_OUT_QCFLUX1LONG_C1,
            site, facility, begin_date_string, end_date_string, datastream_data_out)

        beflux_ds_name = Beflux1long.get_product_name(self.DS_OUT_BEFLUX1LONG_C1, 
            site, facility)
        year, month = begin_date_string[:4], begin_date_string[4:6]
        day = begin_date_string[-2:]
        write_dir = getenv("QUICKLOOK_DATA") + f'/{site}/{beflux_ds_name}/{year}/{month}/{day}'
        if not isdir(write_dir):
            makedirs(write_dir)

        # I don't ever expect there to be more than one file per day
        # but let's put in safeguards regardless
        assert len(beflux_files) == len(qcflux_files)
        
        for i in range(len(beflux_files)):
            beflux1long_out = xr.open_dataset(beflux_files[i])
            qcflux1long_out = xr.open_dataset(qcflux_files[i])

            if len(beflux_files) > 1:
                time_string = beflux_files[i].split('.')[-2]
                begin_date_string += '.' + time_string

            beflux1long_plots(write_dir, beflux_ds_name, begin_date_string, 
                beflux1long_out, qcflux1long_out)


def main():
    sys.exit(Beflux1long().run())


if __name__ == '__main__':
    main()
