import numpy as np
import xarray as xr
from adi_py import SkipProcessingIntervalException
import dsproc3 as dsproc
from typing import Union

try:
    from .var_names import *
    from .instrument_source import *
    from .find_first_or_last_index import *
except ImportError:
    from var_names import *
    from instrument_source import *
    from find_first_or_last_index import *

from pdb import set_trace

class BestEstimate():
    EXTRA_MINUTES_RETRIEVED = 7
    len_input_arrays = 1440 + 2 * EXTRA_MINUTES_RETRIEVED

    sirsC:xr.Dataset = None
    sirsE:xr.Dataset = None
    beflux1long_in:xr.Dataset = None

    data_flag_weekly_last_index = None
    placeholder_var_data = None
    missing_value = -9999
    qc_missing_value = 1

    begin_date:np.datetime64 = None
    end_date:np.datetime64 = None
    begin_input:np.datetime64 = None
    end_input:np.datetime64 = None
    output_start_time = None
    output_end_time = None
    retrieve_qc_vars_after = np.datetime64('2012-02-27', 'ns')
    # keys: name, fill_to, fill_from,
    # replace_through, replace_from, gaps
    # fill_to is how many MISSINGS to insert at beginning
    # fill_from is last valid index before you fill with MISSINGs
    # gaps is a list of (index, length) tuples
    # describing where there are missing samples in the middle of the data
    # replace_through and replace_from are indices to/from which
    # you replace existing data with MISSING
    # to match output_start_time and output_end_time
    # (fill_from, replace_through, gaps, and replace_from are all offset by fill_to)
    dataset_gap_info = {} 


    dataset_to_varname_suffix = {
        'sirsE' : '_E13',
          'brs' : '_brs'
    }

    qc_diff_names = {
        short_direct_normal : 'qcdir',
    down_short_diffuse_hemisp : 'qcdif',
            down_long_hemisp : 'qcdlf',
            up_short_hemisp : 'qcuhf',
                up_long_hemisp : 'qculf'
    }
    
    # For the source_ variables
    source_sirsE_bit = 1
    source_sirsC_bit = 2
    source_brs_bit = 4

    FLOAT_EQ_TOLERANCE=1E-10

    def __init__(self):
        self.input_data_dicts = None

        self.best_estimates:np.ndarray = None
        self.qc_best_estimates:np.ndarray = None
        self.instrument_diffs:np.ndarray = None
        self.data_flag_out:np.ndarray = None
        self.varname:str = None
        self.data_flag_weekly:np.ndarray = None
        self.applied_min_and_max = False
        self.minimum = None
        self.maximum = None
        self.apply_min_and_max_during_best_estimate = False

    ###########################################################################
    # Functions for filling missing timestamps with placeholder data          #
    # so all input arrays have 1440 + (2 * EXTRA_MINUTES_RETRIEVED) samples   #
    ###########################################################################

    def get_dataset_begin_end_fill_index(time_in:np.ndarray) -> tuple:
        begin_gap = (time_in[0] - BestEstimate.begin_input
            ).astype('timedelta64[m]').astype('int32')
        # end_input is one past the last valid timestamp
        end_gap = (BestEstimate.end_input - time_in[-1]
            ).astype('timedelta64[m]').astype('int32') - 1
        begin_index = begin_gap
        end_index = BestEstimate.len_input_arrays - end_gap 
        return begin_index, end_index

    def conform_to_input_start_end_times(data:np.ndarray, replace_through:int,
        replace_from:int, placeholder=missing_value) -> None:
        data[:replace_through] = placeholder
        data[replace_from:] = placeholder

    def fill_begin_and_end_gaps(data:np.ndarray, begin_index:int, 
        end_index:int, placeholder=missing_value) -> np.ndarray:
        begin_filler = np.full(begin_index, placeholder)
        end_filler = np.full(BestEstimate.len_input_arrays - end_index, 
            placeholder)
        return np.concatenate((begin_filler, data, end_filler))

    # Assumption: start and end gaps have already been filled
    # and gap indices are offset by begin_index
    def fill_middle_gaps(data:np.ndarray, gaps:list, 
        placeholder=missing_value) -> np.ndarray:
        if not gaps:
            return data

        result = np.full(BestEstimate.len_input_arrays, placeholder, dtype=data.dtype)
        result_index = 0

        for i in range(len(gaps)):
            gap_index, gap_length_minutes = gaps[i]
            if i == 0:
                chunk_to_copy = data[:gap_index + 1]
                n_vals_to_copy = len(chunk_to_copy)
                result[:n_vals_to_copy] = chunk_to_copy
                result_index += n_vals_to_copy

            result_index += gap_length_minutes

            if i + 1 == len(gaps):
                chunk_to_copy = data[gap_index + 1:]
            else:
                chunk_to_copy = data[gap_index + 1:gaps[i + 1][0] + 1]

            n_vals_to_copy = len(chunk_to_copy)
            result[result_index:result_index + n_vals_to_copy] = chunk_to_copy
            result_index += n_vals_to_copy

        return result

    def fill_gaps(data:np.ndarray, begin_index:int, end_index:int, 
        gap_indices:list, placeholder=missing_value) -> np.ndarray:
        data = BestEstimate.fill_begin_and_end_gaps(data, begin_index, end_index,
            placeholder=placeholder)
        return BestEstimate.fill_middle_gaps(data, gap_indices, placeholder=placeholder)

    def conform_to_input_start_end_times(data:np.ndarray, replace_through:int,
        replace_from:int, placeholder=missing_value) -> None:
        data[:replace_through + 1] = placeholder
        data[replace_from:] = placeholder

    def set_gap_data(dataset:xr.Dataset, dataset_name:str) -> None:
        begin_index, end_index, replace_through, replace_from, gap_indices = \
            BestEstimate.get_gap_data(dataset)
        BestEstimate.dataset_gap_info[dataset_name] = {}
        BestEstimate.dataset_gap_info[dataset_name]['fill_to'] = begin_index
        BestEstimate.dataset_gap_info[dataset_name]['fill_from'] = end_index
        BestEstimate.dataset_gap_info[dataset_name]['replace_through'] = replace_through
        BestEstimate.dataset_gap_info[dataset_name]['replace_from'] = replace_from
        BestEstimate.dataset_gap_info[dataset_name]['gaps'] = gap_indices

    def get_dataset_replace_through_from_index(time_in:np.ndarray) -> tuple:
        # Calling function is responsible for offsetting these
        # by fill_to
        replace_through = find_first_index_where(time_in,
            lambda x: x >= BestEstimate.output_start_time)
        replace_from = find_last_index_where(time_in,
            lambda x: x <= BestEstimate.output_end_time) + 1

        return replace_through, replace_from

    def generate_duplicate_samples_error_message(time_diff:np.ndarray,
        sampling_interval:np.timedelta64, 
        duplicate_sample_indices:np.ndarray=None):
        if duplicate_sample_indices is None:
            duplicate_sample_indices = np.array(time_diff < sampling_interval).nonzero()[0]
        error_message = "found duplicate time samples at the following indices:"
        for index in duplicate_sample_indices:
            error_message += f' ({index},{index + 1})'
        return error_message

    # Throws IndexError if dataset has more than 1454 samples
    def get_gap_data(dataset:xr.Dataset) -> tuple:
        one_minute = np.timedelta64(1, 'm').astype('timedelta64[ns]')
        time = dataset['time'].data
        if len(time) > BestEstimate.len_input_arrays:
            raise IndexError(BestEstimate.generate_duplicate_samples_error_message(
                np.diff(time), one_minute))

        begin_index, end_index = BestEstimate.get_dataset_begin_end_fill_index(time)
        replace_through, replace_from = \
            BestEstimate.get_dataset_replace_through_from_index(time)
        replace_through += begin_index
        replace_from += begin_index

        # If we only have gaps at the beginning and end
        if len(time) + begin_index == end_index:
            gaps = []
        else:
            one_minute = np.timedelta64(1, 'm').astype('timedelta64[ns]')
            time_diff = np.diff(time)
            gap_indices = np.array(time_diff > one_minute).nonzero()[0]
            duplicate_sample_indices = np.array(time_diff < one_minute).nonzero()[0]
            
            if len(duplicate_sample_indices) > 0:
                raise IndexError(BestEstimate.generate_duplicate_samples_error_message(
                    time_diff, one_minute,
                    duplicate_sample_indices=duplicate_sample_indices))

            # List of tuples: (index, length)
            gaps = []
            sum_gap_length = 0
            for gap_index in gap_indices:
                # Subtract 1
                # because a 1-minute difference is considered a gap of 0
                gap_length_minutes = (time[gap_index + 1] - time[gap_index]
                    ).astype('timedelta64[m]').astype('int32') - 1
                gaps.append((gap_index + begin_index, gap_length_minutes))
                sum_gap_length += gap_length_minutes

                # end_index is calculated from len_input_arrays
                # and can play with middle gaps just fine
                # however, replace_from needs to be adjusted
                if gap_index < replace_from:
                    replace_from += gap_length_minutes

            # I think the duplicate_sample_indices check above should always catch this
            # but I'll leave this check in just in case
            if sum_gap_length + len(time) > BestEstimate.len_input_arrays:
                raise IndexError(f'dataset has {len(time)} samples ' +
                    f'(expected {BestEstimate.len_input_arrays})')
                
        return begin_index, end_index, replace_through, replace_from, gaps

    # C VAP sometimes returns fewer than 1440 samples.
    # Although we will always output 1440 samples,
    # we fill with MISSING before/after a certain time,
    # to match the C VAP's output.

    # Correct priority order to mimic C VAP:
    # sirsE, sirsC, brs, sirosedutt, sirsedutt, siroscdutt, sirscdutt, bsrndutt, brsdutt
    # Assumption: we have already verified that at least one of these is not None

    # This function must be called after set_input_data_vars.
    def set_input_start_and_end_timestamps(datasets:list) -> None:
        for dataset in datasets:
            if dataset is not None:
                time_in = dataset['time'].data
                BestEstimate.output_start_time = time_in[0]
                BestEstimate.output_end_time = time_in[-1]

                # to_output_interval needs this
                # in order to know which indices it should overwrite with MISSING
                BestEstimate.adjust_output_interval_begin = (BestEstimate.output_start_time 
                    - BestEstimate.begin_input
                    ).astype('timedelta64[m]').astype('int32') + BestEstimate.EXTRA_MINUTES_RETRIEVED
                BestEstimate.adjust_output_interval_end = (BestEstimate.output_end_time 
                    - BestEstimate.begin_input
                    ).astype('timedelta64[m]').astype('int32') - BestEstimate.EXTRA_MINUTES_RETRIEVED
                return

        raise SkipProcessingIntervalException("No SIRS or BRS data for this time window")

    def set_input_data_vars(begin_date:int, end_date:int) -> None:
        BestEstimate.begin_date = np.datetime64(begin_date, 's').astype('datetime64[ns]')
        BestEstimate.end_date = np.datetime64(end_date, 's').astype('datetime64[ns]')
        extra_time_retrieved = np.timedelta64(BestEstimate.EXTRA_MINUTES_RETRIEVED, 'm')
        BestEstimate.begin_input = BestEstimate.begin_date - extra_time_retrieved
        BestEstimate.end_input = BestEstimate.end_date + extra_time_retrieved

        all_missing = np.full(BestEstimate.len_input_arrays, BestEstimate.missing_value,
            dtype=float)
        all_qc_bad = np.full(BestEstimate.len_input_arrays, BestEstimate.qc_missing_value)
        BestEstimate.placeholder_var_data = { 'data' : all_missing, 'qc' : all_qc_bad, 'source' : 'Not used' }

    def get_longest_time_in(datasets:list, dsids:list) -> int:
        len_input_arrays = 0
        time_in = None
        dsid_time_in = None

        for i in range(len(datasets)):
            dataset = datasets[i]
            if dataset is not None:
                if len(dataset['time'].data) > len_input_arrays:
                    time_in = dataset['time'].data
                    len_input_arrays = len(time_in)
                    if len_input_arrays == BestEstimate.len_input_arrays:
                        return time_in, dsids[i]
                    dsid_time_in = dsids[i]
        return time_in, dsid_time_in

    def set_datasets(sirsC:xr.Dataset, sirsE:xr.Dataset, 
        beflux1long_in:xr.Dataset) -> None:
        BestEstimate.sirsC = sirsC
        BestEstimate.sirsE = sirsE
        BestEstimate.beflux1long_in = beflux1long_in

    def to_output_interval(data:np.ndarray, placeholder=missing_value) -> np.ndarray:
        # We expect that EXTRA_MINUTES_RETRIEVED will never be 0
        data[:BestEstimate.adjust_output_interval_begin] = placeholder
        data[BestEstimate.adjust_output_interval_end + 1:] = placeholder
        return data[BestEstimate.EXTRA_MINUTES_RETRIEVED:
            -BestEstimate.EXTRA_MINUTES_RETRIEVED]

    def set_output_vars(self, ds_out:xr.Dataset) -> None:
        ds_out[self.varname].data = \
            BestEstimate.to_output_interval(self.best_estimates)
        ds_out['qc_' + self.varname].data = \
            BestEstimate.to_output_interval(self.qc_best_estimates,
            placeholder=self.qc_missing_value)
        ds_out[self.varname + '_diff'].data = \
            BestEstimate.to_output_interval(self.instrument_diffs)
        ds_out[self.varname + '_flag'].data = \
            BestEstimate.to_output_interval(self.data_flag_out)

        # Set our source_ variables and populate the bit descriptions
        source_var_name = 'source_' + self.varname
        if source_var_name in ds_out:
            ds_out[source_var_name].data = \
                BestEstimate.to_output_interval(self.source_bits)
            for i in range(len(self.input_dataset_names)):
                ds_out[source_var_name].attrs[f'bit_{i + 1}_description'] = \
                    self.input_data_dicts[self.input_dataset_names[i]]['source']

    def get_input_data_dicts(self) -> list:
        if self.input_data_dicts is None:
            raise Exception("process_data_hook requested best estimate input data " +
                "before calling calc_best_estimate()")
        return list(self.input_data_dicts.values())

    def set_val_lookup_table(self, lookup_table:dict, val1:float, val2:float, val3:float) -> None:
        lookup_table[self.SOURCE_1] = val1
        lookup_table[self.SOURCE_2] = val2
        lookup_table[self.SOURCE_3] = val3

    def get_data_qc_dict(dataset:xr.Dataset, varname:str, dataset_name:str) -> None:
        try:
            varname_suffix = BestEstimate.dataset_to_varname_suffix[dataset_name]
        except KeyError:
            varname_suffix = ''

        varname = varname + varname_suffix

        if dataset is None:
            result = BestEstimate.placeholder_var_data
        else:
            # Our code expects exactly 1440 + (2 * EXTRA_MINUTES_RETRIEVED) samples
            # with all the same timestamps as the other input arrays
            # If we have a late start, early ending, or gaps in the middle
            # we need to fill that with MISSING_VALUE
            if dataset_name not in BestEstimate.dataset_gap_info:
                try:
                    BestEstimate.set_gap_data(dataset, dataset_name)
                except IndexError as e:
                    raise SkipProcessingIntervalException(
                        f'Error reading in data from {dataset_name}: {str(e)}')

            fill_to = BestEstimate.dataset_gap_info[dataset_name]['fill_to']
            fill_from = BestEstimate.dataset_gap_info[dataset_name]['fill_from']
            replace_through = BestEstimate.dataset_gap_info[dataset_name]['replace_through']
            replace_from = BestEstimate.dataset_gap_info[dataset_name]['replace_from']
            gap_indices = BestEstimate.dataset_gap_info[dataset_name]['gaps']

            try:
                data_in = dataset[varname].data

                result = {
                    'data' : BestEstimate.fill_gaps(data_in, fill_to, fill_from, gap_indices)
                }
                BestEstimate.conform_to_input_start_end_times(result['data'],
                    replace_through, replace_from)

                try:
                    # A bunch of input QC vars are all 1's before 2012-02-28
                    # If running before that date, 
                    # skip to "write our own QC data" condition
                    if BestEstimate.begin_date > BestEstimate.retrieve_qc_vars_after:
                        qc_in = dataset['qc_' + varname].data.astype(np.int32)
                        result['qc'] = BestEstimate.fill_gaps(qc_in, fill_to, fill_from,
                            gap_indices, placeholder=BestEstimate.missing_value)
                        BestEstimate.conform_to_input_start_end_times(result['qc'],
                            replace_through, replace_from, 
                            placeholder=BestEstimate.missing_value)
                    else:
                        raise KeyError

                except KeyError:
                    result['qc'] = np.zeros(result['data'].shape, dtype=np.int32)
                    result['qc'][result['data'] < BestEstimate.missing_value + 1] = \
                        BestEstimate.qc_missing_value

                retrieved_var = dsproc.get_retrieved_var(varname, 0)
                result['source'] = dsproc.get_source_ds_name(retrieved_var) + ':' +\
                    dsproc.get_source_var_name(retrieved_var)

            except KeyError:
                result = BestEstimate.placeholder_var_data
        result['suffix'] = varname_suffix
        return result
    
    def last_valid_data_index_from_beflux(self, data_flag_weekly:np.ndarray) -> int:
        if data_flag_weekly is None:
            return None
        for i in range(len(data_flag_weekly) - 1, -1, -1):
            if data_flag_weekly[i] != SOURCE_NONE and \
                data_flag_weekly[i] in self.valid_data_flag_values:
                return i
        return 0

    def get_last_best_estimate_source(self, data_flags:np.ndarray, 
        data_flags_weekly:np.ndarray | None,
        last_best_estimate_index_today:int | None) -> int:
        if last_best_estimate_index_today is not None:
            return int(data_flags[last_best_estimate_index_today])

        # If there is no weekly data, we just consider the check passed
        if data_flags_weekly is None:
            return None

        last_best_estimate_index_weekly = self.last_valid_data_index_from_beflux(data_flags_weekly)
        if last_best_estimate_index_weekly is not None:
            return int(data_flags_weekly[last_best_estimate_index_weekly])

        return SOURCE_NONE

    def get_valid_data_flag_values(split_source_flag_dict:dict) -> list:
        valid_data_flag_values = list(split_source_flag_dict.keys())
        valid_data_flag_values.append(FLAG_INSUFFICIENT_DATA)
        valid_data_flag_values.append(SOURCE_NONE)
        return valid_data_flag_values

    # Return: selected value 1, selected value 2 (or None), source flag
    def select_instrument_source(self, val1:float, val2:float, val3:float,
        data_available_flag:int) -> tuple:
        if data_available_flag == AVAILABLE_NONE:
            return None, None, SOURCE_NONE

        if data_available_flag == AVAILABLE_1:
            return val1, None, self.SOURCE_1
        elif data_available_flag == AVAILABLE_2:
            return val2, None, self.SOURCE_2
        elif data_available_flag == AVAILABLE_3:
            return val3, None, self.SOURCE_3
        elif data_available_flag == AVAILABLE_1_2:
            return val1, val2, self.SOURCE_1_2
        elif data_available_flag == AVAILABLE_1_3:
            return val1, val3, self.SOURCE_1_3
        elif data_available_flag == AVAILABLE_2_3:
            return val2, val3, self.SOURCE_2_3

        else: # three instruments' data available
            diff_1_2 = abs(val1 - val2)
            diff_1_3 = abs(val1 - val3)
            diff_2_3 = abs(val2 - val3)

            smallest_diff = min(diff_1_2, diff_1_3, diff_2_3)

            if smallest_diff == diff_1_2:
                return val1, val2, self.SOURCE_1_2
            elif smallest_diff == diff_2_3:
                return val2, val3, self.SOURCE_2_3
            else:
                return val1, val3, self.SOURCE_1_3

    def get_data_flag_weekly_last_index(data_flag_weekly:np.ndarray) -> np.ndarray:
        return (len(data_flag_weekly) - 
            (BestEstimate.len_input_arrays - 
            BestEstimate.EXTRA_MINUTES_RETRIEVED * 2))
    
    def set_value_failed(self, offset:int) -> None:
        self.best_estimates[offset] = self.missing_value
        self.qc_best_estimates[offset] = self.qc_missing_value
        self.instrument_diffs[offset] = self.missing_value

    def get_weekly_flag(self, varname:str) -> np.ndarray:
        data_flag_weekly = None
        if self.beflux1long_in is not None:
            try:
                data_flag_weekly = self.beflux1long_in[varname + '_flag'].data
                if data_flag_weekly is not None:
                    BestEstimate.get_data_flag_weekly_last_index(data_flag_weekly)
                    if BestEstimate.data_flag_weekly_last_index != -1:
                        data_flag_weekly = data_flag_weekly[:BestEstimate.data_flag_weekly_last_index]
                        if len(data_flag_weekly) == 0:
                            return None

            except KeyError:
                pass

        return data_flag_weekly

    def calc_best_estimate(self) -> None:
        _, self.data_flag_weekly = self.get_input_fields()
        self.init_source_bit_lookup()

        self.best_estimates = np.ndarray(self.len_input_arrays)
        self.qc_best_estimates = np.zeros(self.len_input_arrays, dtype=np.int32)
        self.instrument_diffs = np.ndarray(self.len_input_arrays)
        self.data_flag_out = np.ndarray(self.len_input_arrays)

        last_best_estimate_index = None
        source_flag_to_val = {}


        for i in range(BestEstimate.len_input_arrays):
            val1, val2, val3 = self.get_vals_this_offset(i)
            qc_val1, qc_val2, qc_val3 = self.get_vals_this_offset(i, key='qc')
            if qc_val3 is None:
                qc_val3 = self.qc_missing_value
            self.set_val_lookup_table(source_flag_to_val, val1, val2, val3)

            data_available_flag = get_flag_data_available(qc_val1, qc_val2, qc_val3,
                self.qc_missing_value)

            val_using_1, val_using_2, self.data_flag_out[i] = \
                self.select_instrument_source(val1, val2, val3, data_available_flag)


            if self.data_flag_out[i] == SOURCE_NONE:
                self.set_value_failed(i)
                continue

            if self.should_check_last_best_estimate(val_using_1, val_using_2, i,
                data_available_flag):
                last_best_estimate_source = self.get_last_best_estimate_source(self.data_flag_out,
                    self.data_flag_weekly, last_best_estimate_index)
                if last_best_estimate_source == SOURCE_NONE:
                    self.data_flag_out[i] = SOURCE_NONE
                    self.set_value_failed(i)
                    continue

                # Return of None (no beflux1long_in data) is treated as a pass
                # and value of data_flag_out is unchanged
                elif last_best_estimate_source is not None:
                    self.check_last_best_estimate(i, last_best_estimate_source, data_available_flag)

            if self.data_flag_out[i] == SOURCE_NONE:
                self.set_value_failed(i)

            else:
                last_best_estimate_index = i

                # If we did decide to use both our selected instruments,
                # we don't need to worry about resetting best_estimates or instrument_diffs
                # because those were set in should_check_last_best_estimate
                if self.num_instruments_per_flag(self.data_flag_out[i]) == 1:
                    self.best_estimates[i] = source_flag_to_val[self.data_flag_out[i]]
                    self.instrument_diffs[i] = self.missing_value

            if (self.data_flag_out[i] != SOURCE_NONE and 
                self.data_flag_out[i] != self.missing_value):
                last_best_estimate_index = i

        self.free_unneeded_input_data()
        self.set_source_bits()

    def calc_best_estimate_diffs(self) -> np.ndarray:
        try:
            qc_varname_out = BestEstimate.qc_diff_names[self.varname]
        except KeyError:
            return None

        self.best_estimate_diffs = {}

        best_estimate_bad_mask = np.asarray(self.qc_best_estimates > 0)

        for dataset_name in self.input_data_dicts:
            input_data_dict = self.input_data_dicts[dataset_name]
            input_data = input_data_dict['data']
            qc_input_data = input_data_dict['qc']
            input_bad_mask = np.asarray(
                qc_input_data & BestEstimate.qc_missing_value > 0)
            best_estimate_diff = self.best_estimates - input_data
            best_estimate_diff[np.logical_or(best_estimate_bad_mask, 
                input_bad_mask)] = self.missing_value

            self.best_estimate_diffs[qc_varname_out + dataset_name.lower()] = \
                best_estimate_diff

        return self.best_estimate_diffs

    def set_min_and_max(self, minimum, maximum) -> None:
        self.minimum = minimum
        self.maximum = maximum

    def apply_min_and_max_one_index(self, offset:int) -> None:
        # min is always a scalar; max may or may not be
        if isinstance(self.maximum, np.ndarray):
            max_this_index = self.maximum[offset]
        else:
            max_this_index = self.maximum

        if (self.best_estimates[offset] < self.minimum or
            self.best_estimates[offset] > max_this_index):
            self.best_estimates[offset] = self.missing_value
            self.qc_best_estimates[offset] = self.qc_missing_value
            self.data_flag_out[offset] = self.missing_value

    def set_best_estimate_diffs(self, ds_out:xr.Dataset) -> None:
        for varname_out in self.best_estimate_diffs:
            ds_out[varname_out].data = BestEstimate.to_output_interval(
                self.best_estimate_diffs[varname_out])

    def set_best_estimate_only(self, ds_out:xr.Dataset) -> None:
        ds_out[self.varname].data = \
            BestEstimate.to_output_interval(self.best_estimates)

    def apply_min_and_max(self) -> None:
        var_data = self.best_estimates
        qc_data = self.qc_best_estimates

        bad_value_mask = np.logical_or(np.asarray(var_data > self.maximum),
            np.asarray(var_data < self.minimum))
        var_data[bad_value_mask] = self.missing_value
        self.data_flag_out[bad_value_mask] = self.missing_value
        self.instrument_diffs[bad_value_mask] = self.missing_value
        self.source_bits[bad_value_mask] = 0
        qc_data[bad_value_mask] |= self.qc_missing_value

    # Fulfills a similar role to the _flag var, except:
    # it conforms to current ARM standards; and
    # it says the specific datastream instead of just the retrieval group
    # e.g. sgpsirsC1.a2 instead of just sirsC
    # 
    # Note: when implementing source_bits_lookup,
    # know that instrument flags may be negative
    def set_source_bits(self) -> None:
        self.init_source_bit_lookup()
        self.source_bits = np.zeros(BestEstimate.len_input_arrays, dtype=np.int32)
        no_source = (FLAG_INSUFFICIENT_DATA, SOURCE_NONE, BestEstimate.missing_value)
        for i in range(len(self.source_bits)):
            data_flag = int(self.data_flag_out[i])
            if data_flag not in no_source:
                self.source_bits[i] = self.source_bit_lookup[data_flag]

    # To be called in an object's constructor, after init_split_source_flag_dict
    # Does not populate the lookup values
    # (that's init_source_bit_lookup)
    def allocate_source_bit_lookup(self) -> None:
        source_integer_flags = self.split_source_flag_dict.keys()
        min_flag = min(source_integer_flags)
        max_flag = max(source_integer_flags)
        self.source_bit_lookup = np.zeros(max_flag + 1 + abs(min_flag), dtype=np.int32)

    # The following functions are to be defined
    # by derived classes
    def init_source_bit_lookup(self) -> None:
        pass

    def should_set_last_best_estimate_index(self, offset:int) -> bool:
        return True

    def get_vals_this_offset(self, offset:int, key:str='data') -> float:
        return None

    def get_input_fields(self) -> tuple:
        return (None, None)

    def check_last_best_estimate(self, offset:int,
        last_best_estimate_source:int, data_available_flag:int) -> None: 
        pass

    def should_check_last_best_estimate(self, val1:float,
        val2:float, offset:int, data_available_flag:int) -> bool:
        return False

    def free_unneeded_input_data(self) -> None:
        del self.data_flag_weekly

    def init_split_source_flag_dict(self) -> None:
        pass

    def check_last_best_estimate_one_inst_available(self, last_best_estimate_source:int,
        available_source:int) -> int:
        pass

