import xarray as xr
import numpy as np
from subprocess import run
from typing import List
import matplotlib.pyplot as plt

from pdb import set_trace

try:
    from best_estimate_common import BestEstimate
    from BestEstimateDiffuse import BestEstimateDiffuse
    from BestEstimate3Instruments import BestEstimate3Instruments
    from BestEstimate2Instruments import BestEstimate2Instruments
    from var_names import *
    from instrument_source import FLAG_INSUFFICIENT_DATA, SOURCE_NONE
except ImportError:
    from .best_estimate_common import BestEstimate
    from .BestEstimateDiffuse import BestEstimateDiffuse
    from .BestEstimate3Instruments import BestEstimate3Instruments
    from .BestEstimate2Instruments import BestEstimate2Instruments
    from .var_names import *
    from .instrument_source import FLAG_INSUFFICIENT_DATA, SOURCE_NONE

def plot_best_estimates(ds_out:xr.Dataset, varname:str, ax) -> None:
    plot_standard_variable_data(ds_out, varname, ax)

def plot_qc_diffs(qc_out:xr.Dataset, varname:str, ax) -> None:
    time = qc_out['time'].data
    qc_varname = BestEstimate.qc_diff_names[varname]
    sirsc = qc_out[qc_varname + 'sirsc'].data
    sirse = qc_out[qc_varname + 'sirse'].data
    try:
        brs = qc_out[qc_varname + 'brs'].data
    except KeyError:
        brs = None

    ax.title.set_text("Differences between Best Estimate and Instrument Data")
    plt.setp(ax.get_xticklabels(), rotation=30, horizontalalignment='right')
    ax.set_ylabel(qc_out[qc_varname + 'sirsc'].attrs['units'])
    ax.set_ylim([-20, 20])

    line, = ax.plot(time, sirsc, label='SIRS C1')
    line.set_label('SIRS C1')
    line, = ax.plot(time, sirse, label='SIRS E13')
    line.set_label('SIRS E13')
    if brs is not None:
        line, = ax.plot(time, brs, label='BRS C1')
        line.set_label('BRS C1')
    ax.legend()

def add_data_flag_legend(ax, column1:str, column2:str) -> None:
    # build a rectangle in axes coords
    # https://matplotlib.org/2.0.2/users/text_props.html
    left, width = 0, 1
    bottom, height = 0, 1
    right = left + width
    top = bottom + height
    ax.text(left, bottom - 0.1, column1,
        horizontalalignment='left',
        verticalalignment='top',
        transform=ax.transAxes)
    ax.text(left + 0.5, bottom - 0.1, column2,
        horizontalalignment='left',
        verticalalignment='top',
        transform=ax.transAxes)

def plot_source_flag(ds_out:xr.Dataset, varname:str, ax,
    data_flag_legend_col1:str, data_flag_legend_col2:str) -> None:
    var_long_name = ds_out[varname].attrs['long_name']
    time = ds_out['time'].data
    y_vals = ds_out[varname + '_flag'].data
    ax.title.set_text(var_long_name)
    ax.set_ylabel("Flag value")
    ax.set_ylim([SOURCE_NONE, FLAG_INSUFFICIENT_DATA])
    add_data_flag_legend(ax, data_flag_legend_col1, data_flag_legend_col2)
    ax.scatter(time, y_vals, marker='x')

def get_data_flag_legend_3_inst() -> str:
    column1 = [f'Flag = {BestEstimate3Instruments.SOURCE_1_2} - SIRSE + SIRSC',
        f'Flag = {BestEstimate3Instruments.SOURCE_2_3} - BRS + SIRSC',
        f'Flag = {BestEstimate3Instruments.SOURCE_1_3} - SIRSE + BRS',
        f'Flag = {FLAG_INSUFFICIENT_DATA} - Not enough info.']
    column2 = [f'Flag = {BestEstimate3Instruments.SOURCE_3} - BRS only',
        f'Flag = {BestEstimate3Instruments.SOURCE_1} - SIRSE only',
        f'Flag = {BestEstimate3Instruments.SOURCE_2} - SIRSC only',
        f'Flag = {SOURCE_NONE} - All instruments are down']
    return '\n'.join(column1), '\n'.join(column2)

def get_data_flag_legend_2_inst() -> str:
    column1 = [f'Flag = {BestEstimate3Instruments.SOURCE_1_2} - SIRSE + SIRSC',
        f'Flag = {BestEstimate3Instruments.SOURCE_1} - SIRSE only',
        f'Flag = {BestEstimate3Instruments.SOURCE_2} - SIRSC only']
    column2 = [f'Flag = {FLAG_INSUFFICIENT_DATA} - Not enough info.',
        f'Flag = {SOURCE_NONE} - All instruments are down']
    return '\n'.join(column1), '\n'.join(column2)

def plot_instrument_diffs(ds_out:xr.Dataset, varname:str, ax,
    two_instrument:bool) -> None:
    time = ds_out['time'].data
    y_vals = ds_out[varname + '_diff'].data
    title = "Differences between "
    if two_instrument:
        title += "SIRS E13 and SIRS C1"
    else:
        title += "the two instruments used for averaging"

    ax.title.set_text(title)
    plt.setp(ax.get_xticklabels(), rotation=30, horizontalalignment='right')
    ax.set_ylabel(ds_out[varname].attrs['units'])
    ax.set_ylim([-20, 20])
    ax.plot(time, y_vals)

def plot_standard_variable_data(ds_out:xr.Dataset, varname:str, ax) -> None:
    time = ds_out['time'].data
    var_data_array = ds_out[varname]
    y_values = var_data_array.data
    long_name = var_data_array.attrs['long_name']
    units = var_data_array.attrs['units']

    ax.title.set_text(long_name)
    ax.set_ylabel(units)
    ax.plot(time, y_values)

def plot_source_flag_and_instrument_diffs(ds_out:xr.Dataset, varname:str,
    outfile_prefix:str, two_instrument:bool) -> None:
    fig, (flag_ax, diff_ax) = plt.subplots(2, sharex=True, constrained_layout=True)
    fig.set_size_inches(6, 6)
    #fig.tight_layout(pad=10)

    if two_instrument:
        data_flag_legend_col1, data_flag_legend_col2 = get_data_flag_legend_2_inst()
    else:
        data_flag_legend_col1, data_flag_legend_col2 = get_data_flag_legend_3_inst()

    plot_source_flag(ds_out, varname, flag_ax, data_flag_legend_col1,
        data_flag_legend_col2)
    plot_instrument_diffs(ds_out, varname, diff_ax, two_instrument)
    plt.savefig(f'{outfile_prefix}.{varname}_flag.png')
    plt.close()


def plot_be_and_qc_diffs(ds_out:xr.Dataset, qc_out:xr.Dataset, varname:str,
    outfile_prefix:str) -> None:
    fig, (best_estimate_ax, qc_diff_ax) = plt.subplots(2, layout='constrained', sharex=True)
    plot_best_estimates(ds_out, varname, best_estimate_ax)
    plot_qc_diffs(qc_out, varname, qc_diff_ax)
    plt.savefig(f'{outfile_prefix}.{varname}.png')
    plt.close()

def plot_netrad_and_albedo(ds_out:xr.Dataset, outfile_prefix:str) -> None:
    fig, (netrad_ax, albedo_ax) = plt.subplots(2, layout='constrained', sharex=True)
    plt.setp(albedo_ax.get_xticklabels(), rotation=30, horizontalalignment='right')

    plot_standard_variable_data(ds_out, net_surface_radiation, netrad_ax)
    plot_standard_variable_data(ds_out, albedo, albedo_ax)

    plt.savefig(f'{outfile_prefix}.netrad_albedo.png')
    plt.close()

def plot_down_short_hemisp(ds_out:xr.Dataset, outfile_prefix:str) -> None:
    fig, (ax) = plt.subplots(1, layout='constrained')
    plot_standard_variable_data(ds_out, down_short_hemisp, ax)
    plt.savefig(f'{outfile_prefix}.{down_short_hemisp}.png')
    plt.close()

def make_plots(ds_out:xr.Dataset, qc_out:xr.Dataset, varname:str,
    outfile_prefix:str, two_instrument:bool=False) -> None:
    plot_be_and_qc_diffs(ds_out, qc_out, varname, outfile_prefix)
    plot_source_flag_and_instrument_diffs(ds_out, varname,
        outfile_prefix, two_instrument)

def beflux1long_plots(quicklook_dir:str, ds_name:str, date:str, 
    ds_out:xr.Dataset, qc_out:xr.Dataset) -> None:
    outfile_prefix = f'{quicklook_dir}/{ds_name}.{date}'
    three_instrument_vars = [short_direct_normal, down_short_diffuse_hemisp, down_long_hemisp]
    two_instrument_vars = [up_short_hemisp, up_long_hemisp]

    for varname in three_instrument_vars:
        make_plots(ds_out, qc_out, varname, outfile_prefix)
    for varname in two_instrument_vars:
        make_plots(ds_out, qc_out, varname, outfile_prefix, two_instrument=True)
    plot_netrad_and_albedo(ds_out, outfile_prefix)

if __name__ == '__main__':
    site = 'sgp'
    ds_name = f'{site}beflux1longC1.c1'
    qc_name = f'{site}qcflux1longC1.c1'
    data_home = '/data/home/dev/vap/beflux1long/DATA/tpowell'
    datastream_data = data_home + f'/datastream/{site}'
    quicklook_dir = data_home + f'/www/process/{site}/{ds_name}'
    date = '20130601'

    ds_out = xr.open_dataset(f'{datastream_data}/{ds_name}/{ds_name}.{date}.000000.nc')
    qc_out = xr.open_dataset(f'{datastream_data}/{qc_name}/{qc_name}.{date}.000000.nc')
    beflux1long_plots(quicklook_dir, ds_name, date, ds_out, qc_out)