import numpy as np
import pandas
import math
from datetime import datetime

from pdb import set_trace

julian_2000 = datetime(1999, 12, 31)
 
 #
 # $Id: solarpos.c,v 1.1 2001-06-11 23:47:40 shi Exp $
 # Abstract:
 #      calculates the position of the sun
 # Author:
 #      Nels Larson
 # $Log: not supported by cvs2svn $
 # 2024/12/03 10:27 tpowell
 # Ported to Python. 
 # Only ported calculations used in this VAP,
 # i.e., solar distance and azimuth.
 #
 # Revision 3.1  2001/03/27 17:23:04  shippert
 # OUTDATED
 #
 # Revision 3.0  1997/07/18 15:55:18  d3a230
 # Added rcsid and rcsstate variables and return functions, and set the
 # version number to match Nels's (i.e. 3.0).
 #
 #
 # Revision 1.1  1994/06/23  22:37:41  d3a230
 # Initial revision
 #
 # Revision 1.1  1994/05/06  17:40:35  d3a038
 # Initial revision
 #
 #
 #/
 # "solarpos.c" contains the integer function solarposition() for calculating
 #  the position of the Sun as seen from a place on Earth at a specific time.
 #
 #  Version 3.0 - February 20, 1992.
 #
 #  solarposition() employs the low precision formulas for the Sun's coordinates
 #  given in the "Astronomical Almanac" of 1990 to compute the Sun's apparent
 #  right ascension, apparent declination, altitude, atmospheric refraction
 #  correction applicable to the altitude, azimuth, and distance from Earth.
 #  The "Astronomical Almanac" (A. A.) states a precision of 0.01 degree for the
 #  apparent coordinates between the years 1950 and 2050, and an accuracy of
 #  0.1 arc minute for refraction at altitudes of at least 15 degrees.
 #
 #  The following assumptions and simplifications are made:
 #  -> refraction is calculated for standard atmosphere pressure and temperature
 #     at sea level.
 #  -> diurnal parallax is ignored, resulting in 0 to 9 arc seconds error in
 #     apparent position.
 #  -> diurnal aberration is also ignored, resulting in 0 to 0.02 second error
 #     in right ascension and 0 to 0.3 arc second error in declination.
 #  -> geodetic site coordinates are used, without correction for polar motion
 #     (maximum amplitude of 0.3 arc second) and local gravity anomalies.
 #  -> local mean sidereal time is substituted for local apparent sidereal time
 #     in computing the local hour angle of the Sun, resulting in an error of
 #     about 0 to 1 second of time as determined explicitly by the equation of
 #     the equinoxes.
 #
 #  Right ascension is measured in hours from 0 to 24, and declination in
 #  degrees from 90 to -90.
 #  Altitude is measured from 0 degrees at the horizon to 90 at the zenith or
 #  -90 at the nadir. Azimuth is measured from 0 to 360 degrees starting at
 #  north and increasing toward the east at 90.
 #  The refraction correction should be added to the altitude if Earth's
 #  atmosphere is to be accounted for.
 #  Solar distance from Earth is in astronomical units, 1 a.u. representing the
 #  mean value.
 #
 #  The necessary input parameters are:
 #  -> the date, specified in one of three ways:
 #       1) year, month, day.fraction
 #       2) year, daynumber.fraction
 #       3) days.fraction elapsed since January 0, 1900.
 #  -> site geodetic (geographic) latitude and longitude.
 #
 #  Refer to the function declaration for the parameter type specifications and
 #  formats.
 #
 #  solarposition() returns -1 if an input parameter is out of bounds, or 0 if
 #  values were written to the locations specified by the output parameters.
 #/

 #  Author: Nels Larson
 #          Pacific Northwest Lab.
 #          P.O. Box 999
 #          Richland, WA 99352
 #          U.S.A.
 #/

def solar_position(timestamp:pandas.Timestamp, latitude:float, longitude:float) -> tuple:
    if ((latitude < -90 or latitude > 90) or (longitude < -180 or longitude > 180)):
        raise ValueError("solarposition: Invalid geographic coordinates\n" +
            f'latitude: {latitude}\nlongitude: {longitude}\n')

    MISSING_VALUE = -9999
    MEAN_PRESSURE_SEA_LEVEL = 1013.25
    MEAN_TEMP_SEA_LEVEL = 15
    TWO_PI = math.pi * 2

    #delta_julian_calendar = timestamp - julian_2000
    #days_j2000 = (delta_julian_calendar).days - 1.5
    #if timestamp.year > 2000:
     #   days_j2000 += 1
    #hours = delta_julian_calendar.seconds / 3600
    #day_fraction = hours / 24
    #centuries_j2000 = days_j2000 / 36525.0
    #days_j2000 += day_fraction
    days_j2000 = get_days_j2000(timestamp)
    centuries_j2000 = days_j2000 / 36525.0
    # We work in minutely averages so we don't worry about seconds
    day_fraction = timestamp.hour / 24 + timestamp.minute / (24 * 60)
    days_j2000 += day_fraction
    hours = day_fraction * 24

    # Compute solar position parameters.
    # A. A. 1990, C24.
    mean_anomaly = 357.528 + 0.9856003 * days_j2000
    mean_longitude = 280.460 + 0.9856474 * days_j2000
    mean_anomaly = math.radians(mean_anomaly % 360)
    mean_longitude = math.radians(mean_longitude % 360)
    
    mean_obliquity = math.radians(23.439 - 4.0e-7 * days_j2000)
    ecliptic_long = math.radians((1.915 * math.sin(mean_anomaly)) +
        (0.020 * math.sin(2.0 * mean_anomaly))) + mean_longitude
    solar_distance = 1.00014 - 0.01671 * math.cos(mean_anomaly) - \
              0.00014 * math.cos(2.0 * mean_anomaly)

    # Tangent of ecliptic_long separated into sine and cosine parts for ap_ra.
    right_ascension = math.atan2(math.cos(mean_obliquity) * math.sin(ecliptic_long), 
        math.cos(ecliptic_long))

    # Change range of ap_ra from -pi -> pi to 0 -> 2 pi,
    # then put ap_ra in the range 0 -> 24 hours.
    if right_ascension < 0:
        right_ascension += TWO_PI
    right_ascension = 24 * math.modf(right_ascension / (TWO_PI))[0]

    declination = math.asin(math.sin(mean_obliquity) * math.sin(ecliptic_long))

    # Calculate local mean sidereal time.
    # A. A. 1990, B6-B7.
    # Horner's method of polynomial exponent expansion used for gmst0h.
    gmst0h = 24110.54841 + centuries_j2000 * (8640184.812866 + centuries_j2000 *
        (0.093104 - centuries_j2000 * 6.2e-6))
    # Convert gmst0h from seconds to hours and put in the range 0 -> 24.
    gmst0h = (gmst0h / 3600) % 24
    if gmst0h < 0:
        gmst0h += 24

    # Ratio of lengths of mean solar day to mean sidereal day is 1.00273790934
    # in 1990. Change in sidereal day length is < 0.001 second over a century.
    # A. A. 1990, B6.
    SOLAR_SIDEREAL_DAY_RATIO = 1.00273790934
    length_mean_solar_day = gmst0h + (hours * SOLAR_SIDEREAL_DAY_RATIO) +\
        longitude / 15
    length_mean_solar_day = math.modf(length_mean_solar_day / 24)[0] * 24
    if length_mean_solar_day < 0:
        length_mean_solar_day += 24

    # Calculate local hour angle, altitude, azimuth, and refraction correction.
    # A. A. 1990, B61-B62.
    local_hour_angle = length_mean_solar_day - right_ascension
    # Put hour angle in the range -12 to 12 hours.
    if local_hour_angle < -12:
        local_hour_angle += 24
    elif local_hour_angle > 12:
        local_hour_angle -= 24

    # Convert latitude and local_hour_angle to radians.
    latitude = math.radians(latitude)
    local_hour_angle = local_hour_angle / 12 * math.pi

    cos_declination = math.cos(declination)
    sin_declination = math.sin(declination)
    cos_latitude = math.cos(latitude)
    sin_latitude = math.sin(latitude)
    cos_hour_angle = math.cos(local_hour_angle)

    solar_altitude = math.asin(sin_declination * sin_latitude + 
        cos_declination * cos_hour_angle * cos_latitude)

    cos_altitude = math.cos(solar_altitude)
    # Avoid tangent overflow at altitudes of +-90 degrees.
    # 1.57079615 radians is equal to 89.99999 degrees.
    if abs(solar_altitude) < 1.57079615:
        tan_altitude = math.tan(solar_altitude)
    else:
        tan_altitude = MISSING_VALUE

    solar_altitude = math.degrees(solar_altitude)

    # Compute refraction correction to be added to altitude to obtain actual
    # position.
    # Refraction calculated for altitudes of -1 degree or more allows for a
    # pressure of 1040 mb and temperature of -22 C. Lower pressure and higher
    # temperature combinations yield less than 1 degree refraction.
    # NOTE:
    # The two equations listed in the A. A. have a crossover altitude of
    # 19.225 degrees at standard temperature and pressure. This crossover point
    # is used instead of 15 degrees altitude so that refraction is smooth over
    # the entire range of altitudes. The maximum residual error introduced by
    # this smoothing is 3.6 arc seconds at 15 degrees. Temperature or pressure
    # other than standard will shift the crossover altitude and change the error.
    #
    if solar_altitude < -1 or tan_altitude == MISSING_VALUE:
        refraction_correction = 0
    else:
        if solar_altitude < 19.225:
            refraction_correction = (0.1594 + (solar_altitude) * 
                (0.0196 + 0.00002 * (solar_altitude))) * MEAN_PRESSURE_SEA_LEVEL
            refraction_correction /= (1.0 + (solar_altitude) * 
                (0.505 + 0.0845 * (solar_altitude))) * (273.0 + MEAN_TEMP_SEA_LEVEL)
        
        else:
            refraction_correction = 0.00452 * (
                MEAN_PRESSURE_SEA_LEVEL / (273.0 + MEAN_TEMP_SEA_LEVEL)) / tan_altitude

    solar_altitude += refraction_correction
    return solar_altitude, solar_distance

def get_days_j2000(timestamp:pandas.Timestamp) -> int:
    day_of_year = timestamp.day_of_year
    # Construct Julian centuries since J2000 at 0 hours UT of date,
    # days.fraction since J2000, and UT hours.
    delta_years = timestamp.year - 2000
    delta_days = delta_years * 365 + delta_years // 4 + day_of_year
    if (timestamp.year > 2000):
      delta_days += 1

    return delta_days - 1.5

if __name__ == "__main__":
    timestamp_20241203 = datetime.fromtimestamp(1733264601)
    timestamp_19990101 = datetime(1999, 1, 1, hour=22, minute=23, second=21)
    latitude = 36.605
    longitude = -97.485

    solar_altitude, solar_distance = solar_position(
        timestamp_20241203, latitude, longitude)
    zenith = 90 - solar_altitude
    print(f'{zenith:.6f} {solar_distance:.6f}')
    solar_altitude, solar_distance = solar_position(
        timestamp_19990101, latitude, longitude)
    zenith = 90 - solar_altitude
    print(f'{zenith:.6f} {solar_distance:.6f}')