#!/bin/sh

proctype='vap'
process='beflux1long'
package="$proctype-$process"
execname="beflux1long"

# Print usage

script_name=$0

usage()
{
    cat <<EOM

DESCRIPTION

  Build script used to create a local install of this package. It can also
  be used to perform a staged installation by specifying a destdir, in which
  case the files are copied instead of linked. 

SYNOPSIS

  $script_name [--prefix=path] [--destdir=path] [--conf] [--package] [--env-update]

OPTIONS

  --prefix=path   absolute path to installation directory for conf files
                  and link to the executable
                  default: \$VAP_HOME

  --destdir=path  absolute path prepended to prefix used
                  to perform a staged installation

  --conf          only install the conf files

  --package       only install the python package

  --uninstall     uninstall locally installed package

  --env-update    update conda dependencies listed in the
                  environment.yml file.

  --env-clean    clean/remove conda environment before building it again

 --jenkins-build Building script in Jenkins

  -h, --help      display this help message

EOM
}

# Parse command line

for i in "$@"
do
    case $i in
        --prefix=*)       prefix="${i#*=}"
                          ;;
        --destdir=*)      destdir="${i#*=}"
                          ;;
        --conf)           conf_only=1
                          ;;
        --package)        pkg_only=1
                          ;;
        --env-update)     env_update=1
                          ;;
        --env-clean)      env_clean=1
                          ;;
        --uninstall)      uninstall=1
                          ;;
        --jenkins-build)  jenkins_build=1
                          ;;
        -h | --help)      usage
                          exit 0
                          ;;
        *)                usage
                          exit 1
                          ;;
    esac
done

# Get prefix and install_data from environemnt variables if necessary

if [ ! $prefix ]; then
    if [ $VAP_HOME ]; then
        prefix=$VAP_HOME
    else
        echo "Please pass 'prefix' argument or define VAP_HOME environment variable, exiting"
        usage
        exit 1
    fi
fi

# Functions to echo and run commands

run() {
    echo "> $1"
    $1 || exit 1
}

# Install conf files

if [ ! $pkg_only ] && [ -d conf ]; then

    echo "------------------------------------------------------------------"
    confdir="$destdir$prefix/conf/$proctype"
    if [ ! -d $confdir ]; then
        run "mkdir -p $confdir"
    fi

    files=();
    cd conf
    shopt -s nullglob
    for f in *
    do
        if [ -d $f ] && [ -e $f/$f.process ]; then

            echo "confdir: $confdir/$f"
            run "rm -rf $confdir/$f"

            if [ ! $uninstall ]; then
                if [ $destdir ]; then
                    run "cp -R $f $confdir"
                else
                    pwd=`pwd`
                    run "ln -s $pwd/$f $confdir/$f"
                fi
            fi
        else
            files+=( $f )
        fi
    done

    if [ $files ]; then

        confdir+="/$process"
        echo "confdir: $confdir"
        run "rm -rf $confdir"

        if [ ! $uninstall ]; then

            if [ $destdir ]; then
                run "mkdir -p $confdir"
                for f in ${files[@]}
                do
                    run "cp -R $f $confdir"
                done
            else
                pwd=`pwd`
                run "ln -s $pwd $confdir"
            fi

        fi
    fi

    cd ..
fi

# Activate Conda Environment
conda_source="/apps/base/conda/etc/profile.d/conda.sh"
if [ ! -e "$conda_source" ]; then
    conda_source="/apps/base/mamba/etc/profile.d/conda.sh"
fi

run "source $conda_source"

# VAP-Speific Environment
if [ -e "environment.yml" ]; then

    # echo "environment.yml file found"
    conda_env=$(head -n 1 environment.yml | cut -f2 -d ' ')

    creation_conda_env=$conda_env

    if [ $PRODUCTION = "false" ] || [ $jenkins_build ]; then
        creation_conda_env="dev-$conda_env"
    fi  

    echo "Conda env is: $creation_conda_env"

    # Check if environment already exists
    env_exists=$(conda env list | grep $creation_conda_env)

    # Check if you are already in the environment
    if [[ $PATH != *$creation_conda_env* ]]; then

        if [ $env_clean ]; then
            echo "Cleaning conda envirionment $creation_conda_env"
            run "conda env remove -y -n $creation_conda_env"
            env_exists=""
        fi

        if [ -z "$env_exists" ]; then
            # Create the environment and activate
            echo "Conda env '$creation_conda_env' doesn't exist, creating."
            run "conda env create -n $creation_conda_env -f environment.yml --solver=libmamba"
        fi

        run "conda activate $creation_conda_env"
    fi
# Require environment file
else
    echo "Could not find environment.yml file, required for running"
    exit 1
fi

if [ -z $CONDA_PREFIX ]; then
    echo "Conda activation did not work, exiting"
    exit 1
fi

if [ $env_update ]; then
    run "conda env update --solver=libmamba"
fi

# Set Python Commands
pyprefix=$CONDA_PREFIX
python="$pyprefix/bin/python"
pip="$pyprefix/bin/pip"

# Install Python Package


if [ ! $conf_only ]; then

    # Write wrapper script
    wrapper_file="${process}_wrapper.sh"
    rm -f $wrapper_file
    echo "#!/bin/sh" >> $wrapper_file
    echo "# Activate environment and run VAP" >> $wrapper_file
    echo "source $conda_source"  >> $wrapper_file
    echo "conda activate $conda_env"  >> $wrapper_file
    echo "\$CONDA_PREFIX/bin/$execname \$@" >> $wrapper_file
    run "chmod +x $wrapper_file"

    echo "------------------------------------------------------------------"
    bindir=$destdir$prefix/bin
    if [ ! -d $bindir ]; then
        run "mkdir -p $bindir"
    fi

    run "rm -f $bindir/$execname"

    if [ $uninstall ]; then
        if [ $destdir ]; then
            echo "Cannot uninstall package from root path: $destdir"
            echo "The $package package must be uninstalled manually."
        else
            run "$pip uninstall $package"
        fi
    else
        run "$pip install -e ."
        run "mv $wrapper_file $bindir/$execname"
    fi
fi

exit 0

